# SPDX-FileCopyrightText: Copyright (c) 2024-25, NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .ocr import (
    OCRActor,
    OCRCPUActor,
    OCRGPUActor,
    ocr_page_elements,
)

__all__ = [
    "OCRActor",
    "OCRCPUActor",
    "OCRGPUActor",
    "ocr_page_elements",
]
